%==========================================================================
% Estimation of Floating BG/FG Model by Nonparametric Belief Propagation
% Block-wise appearance model estimation using NBP
% 2011. 2. 13. Suha Kwak, POSTECH.
%==========================================================================



function [Qc] = AppModelEst(I, L, Qc_tp, B, Pc, FGest)
     % [input]
     %   I     : the image obesrvation (the current video frame)
     %   L     : the label map - which pixels would be used
     %   Qc_tp : the temporally propagated block appearance models
     %   B     : the specification of the blocks
     %     - brow     : block height
     %     - bcol     : block width
     %     - bnum_row : the number of blocks at row-direction
     %     - bnum_col : the number of blocks at column-direction
     %     - bnum     : the number of the entire blocks (bnum_row x bnum_col)
     %     - bcoord   : the coordinates of the blocks = [minR, minC, maxR, maxC]
     %     - G        : the pairwise graph structure (a binary matrix)
     %   Pc    : the parameters for nonparametric belief propagation     
     %     - epsilon  : for GMM product sampling
     %     - minpratio: how many pixels are needed to be an observation of a block
     %     - knum     : the (maximum) number of kernel points for KDE (observation)
     %     - kstd     : (minimum) bandwidth of the kernels for KDE (observation)
     %     - temp     : temperature parameter to determine bandwidth
     %     - nghb     : (obs) comparative weight to consider the neighboing blocks
     %   FGest : FG model estimation (true) or BG model estimation (false)
     % [return]
     %   Qc : the appearance models of the blocks;
     %        the marginalized posteriors w.r.t. each RGB random variable
     
     



%% APPEARANCE MODEL AGGRAGATION
Qc = cell(1, B.bnum);
Qc_im = cell(1, B.bnum);

% gathering color (appearance) information --------------------------------
% the set of the observed pixel values at each block
Cset = cell(1, B.bnum);
for bidx = 1 : B.bnum
    % image (color) observations from each of the blocks
	bcoord = B.bcoord(bidx, :);
    lo = L(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4));
    cr = I(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 1);
    cg = I(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 2);
    cb = I(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 3);
    
    % update the sets of motion vectors
    c = [cr(:), cg(:), cb(:)]';
    l = lo(:);
    Cset{bidx} = c(:, l ~= 0);
end


for bidx = 1 : B.bnum    
% observation models (Qc_im) by KDE ---------------------------------------
    pts = zeros(0);
    bws = zeros(0);
    ws = zeros(0);
    
    % observation on the own block region
    if size(Cset{bidx}, 2) > Pc.minpratio * B.brow * B.bcol
        pts = Cset{bidx};
        bw = Pc.kstd * ((B.brow * B.bcol / size(pts, 2)) ^ Pc.temp);
        bws = ones(size(pts)) .* bw;        
        ws = ones(1, size(pts, 2));
    elseif FGest
        % when there is no observation in a block, 
        % we do not make its filtered FG model (w/o considering Qc_tp)
        continue;
    end

    % observation on the four neighboring block regions
    [ridx, cidx] = ind2sub([B.bnum_row, B.bnum_col], bidx);
    for iter = 1 : 4
        nbidx = 0;
        if iter == 1 && ridx > 1
            nbidx = sub2ind([B.bnum_row, B.bnum_col], ridx - 1, cidx);
        elseif iter == 2 && ridx < B.bnum_row
            nbidx = sub2ind([B.bnum_row, B.bnum_col], ridx + 1, cidx);
        elseif iter == 3 && cidx > 1
            nbidx = sub2ind([B.bnum_row, B.bnum_col], ridx, cidx - 1);
        elseif iter == 4 && cidx < B.bnum_col
            nbidx = sub2ind([B.bnum_row, B.bnum_col], ridx, cidx + 1);
        end
        if nbidx == 0
            continue;
        elseif size(Cset{nbidx}, 2) < Pc.minpratio * B.brow * B.bcol
            continue;
        end        
        pts = [pts, Cset{nbidx}];
        bw = Pc.kstd * ((B.brow * B.bcol / size(Cset{nbidx}, 2)) ^ Pc.temp);
        bws = [bws, ones(size(Cset{nbidx})) .* bw];
        ws = [ws, ones(1, size(Cset{nbidx}, 2)) .* Pc.nghb];
    end
    % weight normalization
    ws = ws ./ sum(ws);

    % KDE with the own and neighboring observations
    if ~isempty(pts)
        Qc_im{bidx} = kde(pts, bws, ws);
    end


% filtering by the observation model --------------------------------------
    % no evidences at all
    if isempty(Qc_tp{bidx}) && isempty(Qc_im{bidx})
        % skip, an empty KD
        continue;
    elseif isempty(Qc_tp{bidx})
        if getNpts(Qc_im{bidx}) > Pc.knum * 2
            Qc{bidx} = resample(Qc_im{bidx}, Pc.knum, 'discrete');
        else
            Qc{bidx} = Qc_im{bidx};
        end
    elseif isempty(Qc_im{bidx})
        if getNpts(Qc_tp{bidx}) > Pc.knum * 2
            Qc{bidx} = resample(Qc_tp{bidx}, Pc.knum, 'discrete');
        else
            Qc{bidx} = Qc_tp{bidx};
        end
    else
        Qc{bidx} = kde_prod(Pc.knum, {Qc_tp{bidx}, Qc_im{bidx}}, 'gibbs2', Pc.epsilon);
    end
end



