
% Non-parametric belief propagation code
% Written by Danny Bickson.
% updated: May-2009
% Modified by Nam
% updated: Jan-2011
% Modified by Kwak
% updated: Feb-2011
%
%
% input: G - sqaure matrix size n x n containing standard deviation of
% compatibility. G_ij is standard deviation of compatibility function
% between nodes i and j. Note that unlike LDLC, where the
% decoding matrix H is sparse, we assume that the matrix G is sparse.
% self_pot - an array of kde() objects containing the self potentials. The
% self potentials are Gaussian mixtures
% max_iter - maximal number of iterations
% epsilon - convergence threshold
%
% density - max number of Gaussians in each KDE (Kwak)
% type - option for mixture products in Ihler's KDE toolbox
%        ('epsilon', 'gibbs2', 'exact', and 'import' are possible.)
%        (modify "kde_prod" to make more options of the original one possible.)
%
% output: x - a kde of size nx1 containing the beliefs
%

function [x] = NBP(G, self_pot, rounds, epsilon, density, type)

n = size(G,1); % number of check nodes
assert(size(G,2) == size(G,1));

% messages
M = cell(n,n);      % messages at the current time step
Mp = cell(n,n);     % messages at the previous time step


% initial round (r = 1) ---------------------------------------------------
for i = 1 : n
    % initial outgoing messages = smoothed (by compatibility) belief
    if getNpts(self_pot(i)) > 0
        nb_list = find(G(i, :) ~= 0);
        for j = 1 : length(nb_list)
            Mp{i, nb_list(j)} = integral_ekde(self_pot(i), G(i, nb_list(j)));
        end
    % skip - no observattion (initial belief), no messages
    else
        % Mp{i,j} becomes an empty cell.
    end
end


% iterative update for messages (r > 1) -----------------------------------
for r = 2 : rounds
    for i = 1 : n
        nb_list = find(G(i, :) ~= 0);        
        % calculate messages from the target variable to the neighbors
        for j = 1 : length(nb_list)            
            % incoming messages (previous messages)
            toMul = cell(0);
            cc = 1;
            for k = 1 : length(nb_list)
                if j ~= k && ~isempty(Mp{nb_list(k), i})
                    current = Mp{nb_list(k), i};
                    verify_kde(current);
                    toMul{cc} = current;
                    cc = cc + 1;
                end
            end
            % self potential (observation), if exists
            if getNpts(self_pot(i)) > 0
                toMul{cc} = self_pot(i);
            end

            % outgoing message (current messages - updated)
            if ~isempty(toMul)
                prod = kde_prod(density, toMul, type, epsilon);
                M{i, nb_list(j)} = integral_ekde(prod, G(i, nb_list(j)));
            end
        end
    end
    
    % message update
    Mp = M;
end


% believes for each variable ----------------------------------------------
x = cell(1, n);
for i = 1 : n
    toMul = cell(0);
    cc = 1;    
    nb_list = find(G(:, i) ~= 0);
    % product of incoming messages
    for j = 1 : length(nb_list)
        if ~isempty(M{nb_list(j), i})
            current = M{nb_list(j), i};
            verify_kde(current);
            toMul{cc} = current;
            cc = cc + 1;
        end
    end
    % self potential (observation), if exists
    if getNpts(self_pot(i)) > 0
        toMul{cc} = self_pot(i);
    end
    
    % calculate belief if there is at least one opinion 
    % from neighbors or observation.
    if ~isempty(toMul)
        prod = kde_prod(density, toMul, type, epsilon);
        x{i} = prod;
    end
end

