%==========================================================================
% Estimation of Floating BG/FG Model by Nonparametric Belief Propagation
% Label estimation step: conventional pixel-wise BP
% 2011. 2. 12. Suha Kwak, POSTECH.
% 2011. 2. 23. Modified by Woonhyun Nam, POSTECH
%==========================================================================



function [L] = LabelEst(I, Qc_BG, Qc_FG, O, Qv_BG, Qv_FG, B, Pl)
     % [input]
     %   I     : input image
     %   Qc_BG : block-wise appearance models for background (predicted)
     %   Qc_FG : block-wise appearance models for foreground (predicted)
     %   O     : optical flow map of the current time step
     %   Qv_BG : block-wise motion model for background
     %   Qv_FG : block-wise motion model for foreground
     %   B     : the specification of the blocks
     %   - brow      : block height
     %   - bcol      : block width
     %   - bnum_row  : the number of blocks at row-direction
     %   - bnum_col  : the number of blocks at column-direction
     %   - bnum      : the number of the entire blocks (bnum_row x bnum_col)
     %   - bcoord    : the coordinates of the blocks = [minR, minC, maxR, maxC]
     %   - G         : the pairwise graph structure (a binary matrix)
     %   Pl    : the specification of the blocks
     %   - nrnd      : the number of looopy BP rounds 
     %   - cvar      : bandwidth (variance) for compatibility clique potential
     %   - tempa     : temperature parameter for appearance observation likelihoods
     %   - tempm     : temperature parameter for motion observation likelihoods
     %   - epsc      : the lower bound of compatibility clique potentials
     %   - epso      : the lower bound of observation clique potentials
     % [return]
     %   L           : pixel-wise labels (BG/FG)


     
     


%% PRELIMINARY
% image size
[irow, icol, ichn] = size(I);

% enumerator
BG = 1; FG = 2;
LEFT = 1; RIGHT = 2; UPPER = 3; LOWER = 4;
SAME = 1; DIFF = 2;

% threshold for pixel likelihoods with no associated model 
like_nomodel_app = 256 ^ -3;              % the uniform distribution on the color space
like_nomodel_mot = (irow * icol) ^ -1;    % the uniform distribution on the image size


% compatibility clique potentials =========================================
CmpPot = ones(irow, icol, 4, 2);
    % irow, icol : node indices
    % 4          : four direction (1=left, 2=right, 3=upper, 4=lower)
    % 2          : two cases, whether the two labels are same(1) or not(2)
imdif_LR = sum(((I(:, 2 : icol, :) - I(:, 1 : icol - 1, :)) ./ 255) .^ 2, 3);
imdif_TB = sum(((I(2 : irow, :, :) - I(1 : irow - 1, :, :)) ./ 255) .^ 2, 3);

% compatibility clique potential for the case of two same labels
CmpPot(:, 2 : icol, LEFT, SAME) = exp(-imdif_LR ./ Pl.cvar);      % with the left-neighboring node
CmpPot(:, 1 : icol - 1, RIGHT, SAME) = exp(-imdif_LR ./ Pl.cvar); % with the right-neighboring node
CmpPot(2 : irow, :, UPPER, SAME) = exp(-imdif_TB ./ Pl.cvar);     % with the upper-neighboring node
CmpPot(1 : irow - 1, :, LOWER, SAME) = exp(-imdif_TB ./ Pl.cvar); % with the lower-neighboring node
CmpPot(:, :, :, DIFF) = 1 - CmpPot(:, :, :, SAME);
CmpPot = CmpPot .* (1 - 2 * Pl.epsc) + Pl.epsc;





%% OBSERVATION CLIQUE POTENTIAL
% observation clique potentials
ObsPot = zeros(irow, icol, 2);

% observation likelihood w.r.t. appearance model ==========================
% pixel-wise likelihood map (the third index: BG[1] and FG[2])
plike = zeros(irow, icol, 2);

% block-wise likelihood calculation
plike_BG_parfor = cell(1,B.bnum);
plike_FG_parfor = cell(1,B.bnum);
parfor bidx = 1 : B.bnum
    bcoord = B.bcoord(bidx, :);

    % RGB vectors per each pixel in the block
    cr = I(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 1);
    cg = I(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 2);
    cb = I(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 3);
    c = [cr(:), cg(:), cb(:)]';

% likelihoods w.r.t. BG ---------------------------------------------------
    if isempty(Qc_BG{bidx})
        plike_BG_parfor{bidx} = like_nomodel_app;
    else
        % block appearance model
        bQc = Qc_BG{bidx};
        bplike = evaluate(bQc, c);
        plike_BG_parfor{bidx} = ...
            reshape(bplike, [bcoord(3) - bcoord(1) + 1, bcoord(4) - bcoord(2) + 1]);
    end

% likelihoods w.r.t. FG ---------------------------------------------------
    if isempty(Qc_FG{bidx})
        plike_FG_parfor{bidx} = like_nomodel_app;
    else
        % block appearance model
        bQc = Qc_FG{bidx};
        bplike = evaluate(bQc, c);
        plike_FG_parfor{bidx} = ...
            reshape(bplike, [bcoord(3) - bcoord(1) + 1, bcoord(4) - bcoord(2) + 1]);
    end
end
for bidx = 1 : B.bnum
    bcoord = B.bcoord(bidx, :);
    plike(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), BG) = plike_BG_parfor{bidx};
    plike(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), FG) = plike_FG_parfor{bidx};
end
logplike_app = log(plike);



% observation likelihood w.r.t. motion model ==============================
% pixel-wise likelihood map (the third index: BG[1] and FG[2])
plike = zeros(irow, icol, 2);

% block-wise likelihood calculation
plike_BG_parfor = cell(1,B.bnum);
plike_FG_parfor = cell(1,B.bnum);
parfor bidx = 1 : B.bnum
    bcoord = B.bcoord(bidx, :);

    % RGB vectors per each pixel in the block
    vx = O(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 1);
    vy = O(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 2);    
    v = [vx(:), vy(:)]';

% likelihoods w.r.t. BG ---------------------------------------------------
    if isempty(Qv_BG{bidx})
        plike_BG_parfor{bidx} = like_nomodel_mot;
    else
        % block motion model (convolution)
        pts = getPoints(Qv_BG{bidx});
        bws = getBW(Qv_BG{bidx}) + Pl.conv;
        ws = getWeights(Qv_BG{bidx});
        bQv = kde(pts, bws, ws);
        bplike = evaluate(bQv, v);
        plike_BG_parfor{bidx} = ...
            reshape(bplike, [bcoord(3) - bcoord(1) + 1, bcoord(4) - bcoord(2) + 1]);
    end

% likelihoods w.r.t. FG ---------------------------------------------------
    if isempty(Qv_FG{bidx})
        plike_FG_parfor{bidx} = like_nomodel_mot;
    else
        % block motion model (convolution)
        pts = getPoints(Qv_FG{bidx});
        bws = getBW(Qv_FG{bidx}) + Pl.conv;
        ws = getWeights(Qv_FG{bidx});
        bQv = kde(pts, bws, ws);
        bplike = evaluate(bQv, v);
        plike_FG_parfor{bidx} = ...
            reshape(bplike, [bcoord(3) - bcoord(1) + 1, bcoord(4) - bcoord(2) + 1]);
    end
end
for bidx = 1 : B.bnum
    bcoord = B.bcoord(bidx, :);
    plike(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), BG) = plike_BG_parfor{bidx};
    plike(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), FG) = plike_FG_parfor{bidx};
end
logplike_mot = log(plike);


% observation clique potentials ===========================================
logplike = logplike_app .* Pl.tempa + logplike_mot .* Pl.tempm;
logplike = logplike - median(median(median(logplike)));
ObsPot(:, :, BG) = exp(logplike(:, :, BG)) ./ (exp(logplike(:, :, BG)) + exp(logplike(:, :, FG)));
ObsPot(:, :, FG) = 1- ObsPot(:, :, BG);
ObsPot = ObsPot .* (1 - 2 * Pl.epso) + Pl.epso;





%% PIXEL-WISE BINARY BELIEF PROPAGATION
% initialization ==========================================================
% incoming messages (the third index: BG[1] and FG[2])
%   ex. LRMsg(i, j, :) = a message FROM the left neighbor of x_ij TO x_ij
LRMsg = ones(irow, icol, 2);   % left to right messages
RLMsg = ones(irow, icol, 2);   % right to left messages
TBMsg = ones(irow, icol, 2);   % top to bottom messages
BTMsg = ones(irow, icol, 2);   % bottom to top messages

% initial belief = observation clique potential
Belief = ObsPot;


% loopy belief propagation ================================================
for ridx = 1 : Pl.nrnd
	% messages at the previous time step
    LRMsg_p = LRMsg;
    RLMsg_p = RLMsg;
    TBMsg_p = TBMsg;
    BTMsg_p = BTMsg;

% message update ----------------------------------------------------------
    % left to right messages
    partBelief = Belief(:, 1 : icol - 1, :) ./ RLMsg_p(:, 1 : icol - 1, :);
    LRMsg(:, 2 : icol, BG) = partBelief(:, :, BG) .* CmpPot(:, 1 : icol - 1, RIGHT, SAME) + ...
                             partBelief(:, :, FG) .* CmpPot(:, 1 : icol - 1, RIGHT, DIFF);
    LRMsg(:, 2 : icol, FG) = partBelief(:, :, BG) .* CmpPot(:, 1 : icol - 1, RIGHT, DIFF) + ...
                             partBelief(:, :, FG) .* CmpPot(:, 1 : icol - 1, RIGHT, SAME);
    % right to left messages
    partBelief = Belief(:, 2 : icol, :) ./ LRMsg_p(:, 2 : icol, :);
    RLMsg(:, 1 : icol - 1, BG) = partBelief(:, :, BG) .* CmpPot(:, 2 : icol, LEFT, SAME) + ...
                                 partBelief(:, :, FG) .* CmpPot(:, 2 : icol, LEFT, DIFF);
	RLMsg(:, 1 : icol - 1, FG) = partBelief(:, :, BG) .* CmpPot(:, 2 : icol, LEFT, DIFF) + ...
                                 partBelief(:, :, FG) .* CmpPot(:, 2 : icol, LEFT, SAME);
    % top to bottom messages
	partBelief = Belief(1 : irow - 1, :, :) ./ BTMsg_p(1 : irow - 1, :, :);
    TBMsg(2 : irow, :, BG) = partBelief(:, :, BG) .* CmpPot(1 : irow - 1, :, LOWER, SAME) + ...
                             partBelief(:, :, FG) .* CmpPot(1 : irow - 1, :, LOWER, DIFF);
    TBMsg(2 : irow, :, FG) = partBelief(:, :, BG) .* CmpPot(1 : irow - 1, :, LOWER, DIFF) + ...
                             partBelief(:, :, FG) .* CmpPot(1 : irow - 1, :, LOWER, SAME);
    % bottom to top messages
    partBelief = Belief(2 : irow, :, :) ./ TBMsg_p(2 : irow, :, :);
    BTMsg(1 : irow - 1, :, BG) = partBelief(:, :, BG) .* CmpPot(2 : irow, :, UPPER, SAME) + ...
                                 partBelief(:, :, FG) .* CmpPot(2 : irow, :, UPPER, DIFF);
	BTMsg(1 : irow - 1, :, FG) = partBelief(:, :, BG) .* CmpPot(2 : irow, :, UPPER, DIFF) + ...
                                 partBelief(:, :, FG) .* CmpPot(2 : irow, :, UPPER, SAME);

% belief update -----------------------------------------------------------
    % product of messages and observation clique potential
    Belief = LRMsg .* RLMsg .* TBMsg .* BTMsg .* ObsPot;

    % belief normalization
    sumBelief = sum(Belief, 3);
    Belief(:, :, BG) = Belief(:, :, BG) ./ sumBelief;
    Belief(:, :, FG) = Belief(:, :, FG) ./ sumBelief;
end


% final labeling ==========================================================
L = Belief(:, :, FG) > Belief(:, :, BG);
