%==========================================================================
% Estimation of Floating BG/FG Model by Nonparametric Belief Propagation
% Block-wise motion model estimation using NBP
% 2011. 2. 13. Suha Kwak, POSTECH.
%==========================================================================



function [Qv] = MotModelEst(O, L, B, Pv)
     % [input]
     %   O  : the motion map (optical flows)
     %   L  : the label map - which pixels would be used
     %   B  : the specification of the blocks
     %     - brow     : block height
     %     - bcol     : block width
     %     - bnum_row : the number of blocks at row-direction
     %     - bnum_col : the number of blocks at column-direction
     %     - bnum     : the number of the entire blocks (bnum_row x bnum_col)
     %     - bcoord   : the coordinates of the blocks = [minR, minC, maxR, maxC]
     %     - G        : the pairwise graph structure (a binary matrix)
     %   Pv : the parameters for nonparametric belief propagation
     %     - minpratio: how many pixels are needed to be an observation of a block
     %     - knum     : the number of kernel points for KDE (observation)
     %     - kstd     : bandwidth of the kernels for KDE (observation)
     %     - cstd     : bandwidth of the compatibility potentials for NBP
     %     - maxrnd   : the maximum number of rounds for NBP
     %     - epsilon  : convergence threshold for NBP
     % [return]
     %   Qv : the motion models of the blocks;
     %        the marginalized posteriors w.r.t. each motion random variable



%% MOTION MODEL ESTIMATION
for bidx = 1 : B.bnum
    % gathering motion information --------------------------------------------
    bcoord = B.bcoord(bidx, :);
    
    % motion observations from each of the blocks
    lo = L(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4));
    vx = O(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 1);
    vy = O(bcoord(1) : bcoord(3), bcoord(2) : bcoord(4), 2);
    
    % the set of the associated motion vectors
    v = [vx(:), vy(:)]';
    l = lo(:);
    Vset = v(:, l ~= 0);
    
    % observation clique potentials -------------------------------------------
    % sufficient observations - setting the self potential
    if size(Vset, 2) > Pv.minpratio * length(l);
        obsclq(bidx) = kde(Vset, Pv.kstd);
        % insufficient observations
    else
        % ignored (an empty KDE)
        obsclq(bidx) = kde;
    end
end




%% INFERENCE
% compatibility potentials for all of the pairwise cliques
Gv = B.G .* Pv.cstd;

% no BP inference, evidences only =========================================
if sum(sum(Gv)) == 0
    Qv = cell(1, B.bnum);
    for bidx = 1 : B.bnum
        if getNpts(obsclq(bidx)) > 0
            Qv{bidx} = obsclq(bidx);
        end
    end

% appearance model estimation by NBP ======================================
else
    Qv = NBP(Gv, obsclq, Pv.maxrnd, Pv.epsilon, Pv.knum, 'gibbs2');
end



