%==========================================================================
% Estimation of Floating BG/FG Model by Nonparametric Belief Propagation
% Temporal model propagation to bridge motion models and appearance models
% 2011. 2. 14. Suha Kwak, POSTECH.
%==========================================================================



function [Qc_tp] = TempModelPropagate(Qv, Qc_prev, B, Pt)
     % [input]
     %   Qv      : the current block motion models 
     %   Qc_prev : the block appearance models at the previous time step
     %   B       : the specification of the blocks
     %     - brow     : block height
     %     - bcol     : block width
     %     - bnum_row : the number of blocks at row-direction
     %     - bnum_col : the number of blocks at column-direction
     %     - bnum     : the number of the entire blocks (bnum_row x bnum_col)
     %     - bcoord   : the coordinates of the blocks = [minR, minC, maxR, maxC]
     %     - bstd_row : bandwidth of normal kernel approximating block region 
     %     - bstd_col : bandwidth of normal kernel approximating block region 
     %     - G        : the pairwise graph structure (a binary matrix)
     %   Pt      : the set of parameters for temporal model propatation
     %     - browstd  : bandwidth of normal kernel approximating block region 
     %     - bcolstd  : bandwidth of normal kernel approximating block region 
     %     - transtd  : BW of the state transition probability (Gaussian conv.)
     %     - minAW    : the minimum aggregation weights (temporal model propagation)
     % [return]
     %   Qc_tp   : the temporally propagated block appearance models





%% TEMPORAL MODEL PROPAGATION
% block appearance models by temporal model propagation
Qc_tp = cell(1, B.bnum);

for tidx = 1 : B.bnum
    
% preparation -------------------------------------------------------------
    % no temporal model propagation when the block is on the scene boundary
    % (the block is considered pure BGs)
    [ridx, cidx] = ind2sub([B.bnum_row, B.bnum_col], tidx);
    if ridx == 1 || ridx == B.bnum_row || cidx == 1 || cidx == B.bnum_col
        continue;
    end

    % motion model of the target block
    if isempty(Qv{tidx})
        continue;       % skip when there is no motion model estimated
    end
    bQv = Qv{tidx};

    % (center) position of the target block
    bpos = [(B.bcoord(tidx, 2) + B.bcoord(tidx, 4)) / 2 ; ...
            (B.bcoord(tidx, 1) + B.bcoord(tidx, 3)) / 2];
    
    % motion model information
    npts = getNpts(bQv);                            % the # of points
    pts = repmat(bpos, 1, npts) - getPoints(bQv);   % the points
    bws = getBW(bQv);                               % bandwidths of the points
    ws = getWeights(bQv);                           % weights of the points

    % motion model convolution (considering block region)
    bws = bws + repmat([Pt.bcolstd ; Pt.browstd], 1, npts);

    
% block-wise aggregation-weight map ---------------------------------------
    breg_weight = zeros(B.bnum, 1);
    for pidx = 1 : npts
        intg_col = normcdf(B.bcoord(:, [4, 2]), pts(1, pidx), bws(1, pidx));
        intg_row = normcdf(B.bcoord(:, [3, 1]), pts(2, pidx), bws(2, pidx));
        breg_weight = breg_weight + ...
                      ((intg_col(:, 1) - intg_col(:, 2)) .* ...     % column-wise integration
                       (intg_row(:, 1) - intg_row(:, 2)) .* ...     % row-wise integration
                        ws(pidx));                                  % weighting
    end    
    % aggregation-weight thresholding and normalization
    breg_weight(breg_weight < Pt.minAW) = 0;
    breg_weight = breg_weight ./ sum(breg_weight);
    
    
% aggregation of the previous appearance models ---------------------------
    % previous blocks with nonzero aggregation-weight (overlapped blocks)
    oblist = find(breg_weight > 0);

    % (KDE) spec. of the temporally propagated model
    tp_pts = zeros(0);
    tp_bws = zeros(0);
    tp_ws = zeros(0);
    
    % aggregation of the appearance models at the previous time step
    for idx = 1 : length(oblist)
        obidx = oblist(idx);
        if isempty(Qc_prev{obidx})
            continue;
        end
        tp_pts = [tp_pts, getPoints(Qc_prev{obidx})];
        tp_bws = [tp_bws, getBW(Qc_prev{obidx})];
        tp_ws = [tp_ws, getWeights(Qc_prev{obidx}) .* breg_weight(obidx)];
    end
    if isempty(tp_pts)
        continue;
    end
    tp_ws = tp_ws ./ sum(tp_ws);

    % state transition - Gaussian convolution
    tp_bws = tp_bws + Pt.transtd;

    % temporally propagated appearance model
	Qc_tp{tidx} = kde(tp_pts, tp_bws, tp_ws);
end



